/* ********************************************************************** *
   File: knapsack_bsp.c
   Date: 19/04/99
   Description: Knapsack 0-1 problem iplementation using the BSP library.
     Pipeline parallelization technique is used to program this Dinamic 
     Programmation algorithm.
     Each stage of the computation corresponds to the set of decision 
     problems about the inclusion of the i-th object in the knapsack for
     all possible capacity values. The optimal partial solutions to this 
     set of problems are obteined using the optimal partial solutions to 
     the set of problems in the previous stage.
     Each stage is assigned to one processor into the pipeline, and the 
     final optimal solution will be in the processor that computes the last 
     stage.
   Dependences:
    mytime.h : Time measure routines.
   History:
     knapsack_bsp.c: 
 * ********************************************************************** */

#define NEXT          ((Name+1)%Workers)
#define PREVIOUS      ((Workers+Name-1)%Workers)
#define max(a,b)      (((a)>(b))?(a):(b))

#include "bsp.h"
#include "mytime.h"
#include <stdio.h>
#include <stdlib.h>

/* ---------------------- Buffer declaration --------------------------- */
int *buffer, front, end;

/* ********************************************************************* *
   Routine: profit
   Description: Profit function. Returns the profit of the i-th object.
 * ********************************************************************* */
int profit(int i) {
  return i*i;
}

/* ********************************************************************* *
   Routine: weigth
   Description: Weigth function. Returns the weigth of the i-th object.
 * ********************************************************************* */
int weigth(int i) {
  return i;
}

/* ********************************************************************* *
   Routine: work_distribution
   Description: Distributes the objects of decision among available 
     processors using a ciclic assignament. It also determines the proessor 
     that will contain the final solution.
   Parameters:
     int Objects, number of objects
     int Workers, number of processor in the pipeline
     int Name, processor running this routine
     int *Name_Sol, processor that will have the solution
     int *Num_Bands, number of objects assigned to processor Name
 * ********************************************************************* */
void work_distribution(int Objects, int Workers, int Name, int *Name_Sol,
                       int *Num_Bands){
  int Rest=Objects%Workers;
  *Name_Sol=(Rest)?(Rest-1):(Workers-1);
  *Num_Bands=Objects/Workers;
  if(Rest) (*Num_Bands)++;
}

/* ********************************************************************* *
   Routine: dummy_stage
   Description: Perform the necessary dummy steps for processors in the 
     last band.
   Parameters:
     int Capacity, maximum avalilable capacity
     int Workers, number of processors in the pipeline
 * ********************************************************************* */
void dummy_stage(int Capacity, int Workers) {
  int step;
  for(step=0; step<=Capacity+Workers; step++) bsp_sync();
}

/* ********************************************************************* *
   Routine: knapsack_spaa0
   Description: It computes the set of optimal solutions for the set of 
     knapsack problems with Name+1 objects and the capacity value varying
     from 0 to Capacity. It also deals with the buffer management.
   Parameters:
     int Capacity, maximum available capacity
     int Name, processor running this routine (used by macro NEXT)
     int Workers, number of processors in the pipeline (used by macro NEXT)
     int bLast, TRUE if Name is the last object; FALSE otherwise
 * ********************************************************************* */
int knapsack_spaa0(int Capacity, int Name, int Workers, int bLast) {
  int *Q, q, c, W, P, status, step;

  /* Partial solution vector allocation and initialization */
  Q=(int*)calloc(Capacity+1, sizeof(int));
  if(Q==NULL)
    bsp_abort("Memory Error: Processor %d can't allocate %d items\n",
              Name, (Capacity+1)*sizeof(int));

  /* Initialize weigth and profit for object Name+1 */
  W=weigth(Name+1);
  P=profit(Name+1);

  /* Each step is a Super-step */
  for(step=0, c=0; step<=Capacity+Workers; step++) {
    /* Get values from previous stage and save them in the buffer */
    bsp_get_tag(&status, &q);
    bsp_move(NULL, 0);
    if(status!=-1) {
      end=(++end)%(Capacity+1);
      buffer[end]=q;
    }

    /* Load pipeline and solve subproblems */
    if((bsp_pid()<=step) && (c<=Capacity)) {
      /* Get value from buffer */
      q=buffer[front];
      front=(++front)%(Capacity+1);

      /* Update dependencies from q=Opt(Name,c) */
      Q[c]=max(Q[c], q);
      if(c+W<=Capacity)
        Q[c+W]=max(Q[c+W], q+P);

      /* Communication */
      if(!bLast) bsp_send(NEXT, &Q[c], NULL, 0);

      /* Next value for capacity subproblem */
      c++;
    }
    /* Synchronization */
    bsp_sync();
  }

  /* Memory deallocation and return */
  q=Q[Capacity];
  free(Q);
  return q;
}

/* ********************************************************************* *
   Routine: knapsack_spaa
   Description: It computes the set of  optimal solutions for the set of
     knapsack problems with Name+1 objects and capacity value varying
     from 0 to Capacity.
   Parameters:
     int Capacity, maximum available capacity
     int Name, processor running this routine (used by macro NEXT)
     int Workers, number of processors in the pipeline (used by macro NEXT)
     int bLast, TRUE if Name is the last object; FALSE otherwise
 * ********************************************************************* */
int knapsack_spaa(int Capacity, int Name, int Workers, int bLast) {
  int *Q, q, c, W, P, status, step;

  /* Partial solution vector allocation and initialization */
  Q=(int*)calloc(Capacity+1, sizeof(int));
  if(Q==NULL)
    bsp_abort("Memory Error: Processor %d can't allocate %d items\n",
              Name, (Capacity+1)*sizeof(int));

  /* Initialize weigth and profit for object Name+1 */
  W=weigth(Name+1);
  P=profit(Name+1);

  /* Each step is a Super-step */
  for(step=0, c=0; step<=Capacity+Workers; step++) {
    /* Load pipeline and solve subproblems */
    if((bsp_pid()<=step) && (c<=Capacity)) {
      /* Get value from previous stage */
      bsp_get_tag(&status, &q);
      bsp_move(NULL,0);
      if(status==-1)
        bsp_abort("Pipe Error: Step %d Process %d c %d\n", step, Name, c);

      /* Update dependencies from q=Opt(Name,c) */
      Q[c]=max(Q[c], q);
      if(c+W<=Capacity)
        Q[c+W]=max(Q[c+W], q+P);

      /* Communication */
      if(!bLast) bsp_send(NEXT, &Q[c], NULL, 0);

      /* Next value for capacity subproblem */
      c++;
    }
    /* Synchronization */
    bsp_sync();
  }

  /* Memory deallocation */
  q=Q[Capacity];
  free(Q);
  return q;
}

/* ********************************************************************* *
   Routine: main
   Description: Main program
   Usage:
          mpirun -n $Procs knapsack_bsp $Objects $Capacity
 * ********************************************************************* */
int main(int argc, char *argv[]) {
  int Name, Workers, NAME, b, Sol, Objects, Capacity, Num_Bands, Name_Sol;
  int tagsize=sizeof(int);
  double total_time;
  struct timeval start_timer, end_timer;

  /* ------------------ Step=-1: MPI inicialization -------------------- */
  bsp_begin(bsp_nprocs());  
  Workers=bsp_nprocs();
  Name=bsp_pid();

  /* Command line parameters processing */
  if(argc!=3) {
    printf("\nError: Invalid number of parameters (%d)\n", argc);
    printf("Usage: bsprun -n $Procs knapsack_bsp $Objects $Capacity\n");
    exit(1); 
  }
  Objects=atoi(argv[1]);
  Capacity=atoi(argv[2]);

  /* Compute number of Bands and last processor */
  work_distribution(Objects, Workers, Name, &Name_Sol, &Num_Bands);

  /* Buffer allocation and initialization */
  if(Name==0) {
    buffer=(int*)calloc(Capacity+1, sizeof(int));
    if(buffer==NULL)
      bsp_abort("Memory Error: Processor %d can't allocate %d items\n",
                Name, (Capacity+1)*sizeof(int));
    front=0;
    end=Capacity;
  }

  /* Tags size is fixed */
  bsp_set_tagsize(&tagsize);
  bsp_sync();

  /* ------------------- Step=0. Starting pipeline ---------------------- */
  /* Start timer */
  GETTIME(start_timer);

  /* Each processor performs the correspondig bands */
  for(b=0; b<Num_Bands; b++) {
    NAME=b*Workers+Name;
    if(Name==0)
      Sol=knapsack_spaa0(Capacity, NAME, Workers, NAME==Objects-1);
    else if(NAME<Objects)
      Sol=knapsack_spaa(Capacity, NAME, Workers, NAME==Objects-1);
    else
      dummy_stage(Capacity, Workers);
  }

  /* Stop timer and get time */
  GETTIME(end_timer);
  DIFTIME(end_timer, start_timer, total_time);

  /* Only the processor with the optimal solution prints */
  if(Name==Name_Sol) {
   /* printf("\nProc.\tObjects\tBandas\Ult.\tCapacity\tSoluc.\tTiempo");*/
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%.2f\n", Workers, Objects, Num_Bands, 
            Name_Sol, Capacity, Sol, total_time); 
  }
  fflush(stdout);

  /* Buffer deallocation */
  if(Name==0) free(buffer);
  bsp_set_tagsize(&tagsize);
  bsp_end();
  return 0;
}














